from .tclab import TCLab, TCLabModel
from .historian import Historian, Plotter
from .labtime import labtime, clock


class Experiment:
    """Utility class for running experiments on the TCLab

    The class is designed to be used as a context. On initialisation
    it will automatically connect to the TCLab, create a Historian and Plotter
    and provide a `.clock` method.

    >>> with Experiment(connected=False, plot=False, time=20) as experiment:
    ...     for t in experiment.clock():
    ...         experiment.lab.Q1(100 if t < 100 else 0)
    ... # doctest: +SKIP
    ... # doctest: +ELLIPSIS
    TCLab version ...
    Simulated TCLab
    TCLab Model disconnected successfully.

    Once the experiment has run, you can access `experiment.historian` to see
    the results from the simulation.

    """

    def __init__(self, connected=True, plot=True,
                 twindow=200, time=500,
                 dbfile=':memory:',
                 speedup=1, synced=True, tol=0.5):
        """Parameters:
        connected: If True, connect to a physical TCLab, if False, connecte to
                   TCLabModel
        plot: Use a Plotter
        twindow: (only applicable if plotting) the twindow for the Plotter
        time: total time to run for (used for experiment.clock)
        dbfile: The dbfile to use for the Historian
        speedup: speedup factor to use if not connected.
        synced: Try to run at a fixed factor of real time. If this is False, run
                as fast as possible regardless of the value of speedup.
        tol: Clock tolerance (used for experiment.clock)
        """
        if (speedup != 1 or not synced) and connected:
            raise ValueError('The real TCLab can only run real time.')

        self.connected = connected
        self.plot = plot
        self.twindow = twindow
        self.time = time
        self.dbfile = dbfile
        self.speedup = speedup
        self.synced = synced
        self.tol = tol
        if synced:
            labtime.set_rate(speedup)

        self.lab = None
        self.historian = None
        self.plotter = None

    def __enter__(self):
        if self.connected:
            self.lab = TCLab()
        else:
            self.lab = TCLabModel(synced=self.synced)
        self.historian = Historian(self.lab.sources, dbfile=self.dbfile)
        if self.plot:
            self.plotter = Plotter(self.historian, twindow=self.twindow)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lab.close()
        self.historian.close()

    def clock(self):
        if self.synced:
            times = clock(self.time, tol=self.tol)
        else:
            times = range(self.time)
        for t in times:
            yield t
            if self.plot:
                self.plotter.update(t)
            else:
                self.historian.update(t)
            if not self.synced:
                self.lab.update(t)


def runexperiment(function, *args, **kwargs):
    """Simple wrapper for Experiment which builds an experiment and calls a
    function in a timed for loop.

    The function will be passed the time and a TCLab instance at every tick of
    the clock.

    The remaining arguments are passed Experiment.
    """
    with Experiment(*args, **kwargs) as experiment:
        for t in experiment.clock():
            function(t, experiment.lab)
    return experiment



import time as time


class Labtime():
    def __init__(self):
        self._realtime = time.time()
        self._labtime = 0
        self._rate = 1
        self._running = True
        self.lastsleep = 0

    @property
    def running(self):
        """Returns variable indicating whether labtime is running."""
        return self._running

    def time(self):
        """Return current labtime."""
        if self.running:
            elapsed = time.time() - self._realtime
            return self._labtime + self._rate * elapsed
        else:
            return self._labtime

    def set_rate(self, rate=1):
        """Set the rate of labtime relative to real time."""
        if rate <= 0:
            raise ValueError("Labtime rates must be positive.")
        self._labtime = self.time()
        self._realtime = time.time()
        self._rate = rate

    def get_rate(self):
        """Return the rate of labtime relative to real time."""
        return self._rate

    def sleep(self, delay):
        """Sleep in labtime for a period delay."""
        self.lastsleep = delay
        if self._running:
            time.sleep(delay / self._rate)
        else:
            raise RuntimeWarning("sleep is not valid when labtime is stopped.")

    def stop(self):
        """Stop labtime."""
        self._labtime = self.time()
        self._realtime = time.time()
        self._running = False

    def start(self):
        """Restart labtime."""
        self._realtime = time.time()
        self._running = True

    def reset(self, val=0):
        """Reset labtime to a specified value."""
        self._labtime = val
        self._realtime = time.time()


labtime = Labtime()


# for backwards compatability
def setnow(tnow=0):
    labtime.reset(tnow)


def clock(period, step=1, tol=float('inf'), adaptive=True):
    """Generator providing time values in sync with real time clock.

    Args:
        period (float): Time interval for clock operation in seconds.
        step (float): Time step.
        tol (float): Maximum permissible deviation from real time.
        adaptive (Boolean): If true, and if the rate != 1, then the labtime
            rate is adjusted to maximize simulation speed.

    Yields:
        float: The next time step rounded to nearest 10th of a second.


    Note:
        * Passing `tol=float('inf')` will effectively disable sync error checking
        * When large values for `tol` are used, no guarantees are made that the
          last time returned will be equal to `period`.
    """
    start = labtime.time()
    now = 0

    while round(now, 0) <= period:
        yield round(now, 2)
        if round(now) >= period:
            break
        elapsed = labtime.time() - (start + now)
        rate = labtime.get_rate()
        if (rate != 1) and adaptive:
            if elapsed > step:
                labtime.set_rate(0.8 * rate * step / elapsed)
            elif (elapsed < 0.5 * step) & (rate < 50):
                labtime.set_rate(1.25 * rate)
        else:
            if elapsed > step + tol:
                message = ('Labtime clock lost synchronization with real time. '
                           'Step size was {} s, but {:.2f} s elapsed '
                           '({:.2f} too long). Consider increasing step.')
                raise RuntimeError(message.format(step, elapsed, elapsed-step))
        labtime.sleep(step - (labtime.time() - start) % step)
        now = labtime.time() - start



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division
import bisect
import sqlite3
from .labtime import labtime
import time


class TagDB:
    """Interface to sqlite database containing tag values"""
    def __init__(self, filename=":memory:"):
        """Create or connect to a database

        :param filename: The filename of the database.
                         By default, values are stored in memory."""
        self.db = sqlite3.connect(filename)
        self.cursor = self.db.cursor()
        creates = ["""CREATE TABLE IF NOT EXISTS tagvalues (
                           session_id REFERENCES session (id),
                           timeseconds, name, value)""",
                   """CREATE TABLE IF NOT EXISTS sessions (
                           id INTEGER PRIMARY KEY,
                           starttime)"""]
        for statement in creates:
            self.cursor.execute(statement)
        self.db.commit()
        self.session = None

    def new_session(self):
        self.cursor.execute("""INSERT INTO SESSIONS (starttime)
                               VALUES (datetime('now'))""")
        self.session = self.cursor.lastrowid
        self.db.commit()

    def get_sessions(self):
        query = """SELECT id, starttime, COUNT(DISTINCT timeseconds) 
                   FROM sessions LEFT JOIN tagvalues ON sessions.id=session_id
                   GROUP BY id ORDER BY starttime"""
        return list(self.cursor.execute(query))

    def delete_session(self, session_id):
        queries = ['DELETE FROM sessions WHERE id = ?',
                   'DELETE FROM tagvalues WHERE session_id = ?']
        for query in queries:
            self.cursor.execute(query, (session_id,))
        self.db.commit()

    def record(self, timeseconds, name, value):
        if self.session is None:
            self.new_session()
        self.cursor.execute("INSERT INTO tagvalues VALUES (?, ?, ?, ?)",
                            (self.session, timeseconds, name, value))
        self.db.commit()

    def get(self, name, timeseconds=None, session=None):
        if session is None:
            session = self.session
        query = """SELECT timeseconds, value FROM tagvalues
                   WHERE session_id=? AND name=?"""
        parameters = [session, name]
        if timeseconds is not None:
            query += " and timeseconds=?"
            parameters.append(timeseconds)
        query += " ORDER BY timeseconds"
        return list(self.cursor.execute(query, parameters))

    def clean(self):
        """Delete sessions with no associated points"""
        query = """DELETE FROM sessions WHERE id NOT IN 
                   (SELECT DISTINCT session_id FROM tagvalues)"""
        self.cursor.execute(query)
        self.db.commit()

    def close(self):
        self.clean()
        self.db.close()


class Historian(object):
    """Generalised logging class"""
    def __init__(self, sources, dbfile=":memory:"):
        """
        sources: an iterable of (name, callable) tuples
            - name (str) is the name of a signal and the
            - callable is evaluated to obtain the value.

        Example:

        >>> a = 1
        >>> def getvalue():
        ...     return a
        >>> h = Historian([('a', getvalue)])
        >>> h.update(0)
        >>> h.log
        [(0, 1)]


        Sometimes, multiple values are obtained from one function call. In such
        cases, names can still be specified as before, but callable can be
        passed as None for the subsequent names which come from a previous
        callable.

        Example:

        >>> a = 1
        >>> b = 2
        >>> def getvalues():
        ...     return [a, b]
        >>> h = Historian([('a', getvalues),
        ...                ('b', None)])
        >>> h.update(0)
        >>> h.log
        [(0, 1, 2)]

        """
        self.sources = [('Time', lambda: self.tnow)] + list(sources)
        if dbfile:
            self.db = TagDB(dbfile)
            self.db.new_session()
            self.session = self.db.session
        else:
            self.db = None
            self.session = 1

        self.tstart = labtime.time()

        self.columns = [name for name, _ in self.sources]

        self.build_fields()

    def build_fields(self):
        self.fields = [[] for _ in self.columns]
        self.logdict = dict(zip(self.columns, self.fields))
        self.t = self.logdict['Time']

    def update(self, tnow=None):
        if tnow is None:
            self.tnow = labtime.time() - self.tstart
        else:
            self.tnow = tnow

        for name, valuefunction in self.sources:
            if valuefunction:
                v = valuefunction()
                try:
                    values = iter(v)
                except TypeError:
                    values = iter([v])
            try:
                value = next(values)
            except StopIteration:
                raise ValueError("valuefunction did not return enough values")

            self.logdict[name].append(value)
            if self.db and name != "Time":
                self.db.record(self.tnow, name, value)

    @property
    def log(self):
        return list(zip(*[self.logdict[c] for c in self.columns]))

    def timeindex(self, t):
        return max(bisect.bisect(self.t, t) - 1, 0)

    def timeslice(self, tstart=0, tend=None, columns=None):
        start = self.timeindex(tstart)
        if tend is None:
            stop = len(self.t) + 1
        # Ensure that we always return at least one time's value
        if tend == tstart:
            stop = start + 1
        if columns is None:
            columns = self.columns
        return [self.logdict[c][start:stop] for c in columns]

    def at(self, t, columns=None):
        """ Return the values of columns at or just before a certain time"""
        return [c[0] for c in self.timeslice(t, t, columns)]

    def after(self, t, columns=None):
        """ Return the values of columns after or just before a certain time"""
        return self.timeslice(t, columns=columns)

    def _dbcheck(self):
        if self.db is None:
            raise NotImplementedError("Sessions not supported without dbfile")
        return True

    def new_session(self):
        self._dbcheck()
        self.db.new_session()
        self.session = self.db.session
        self.tstart = labtime.time()
        self.build_fields()

    def get_sessions(self):
        self._dbcheck()
        return self.db.get_sessions()

    def load_session(self, session):
        self._dbcheck()
        self.db.session = session
        self.build_fields()
        # FIXME: The way time is handled here is a bit brittle
        first = True
        for name in self.columns[1:]:
            for t, value in self.db.get(name):
                self.logdict[name].append(value)
                if first:
                    self.t.append(t)
            first = False

    def close(self):
        if self.db:
            self.db.close()

    def to_csv(self, filename):
        """Output contents of log file to CSV"""
        import csv

        with open(filename, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.columns)
            writer.writerows(self.log)


class Plotter:
    def __init__(self, historian, twindow=120, layout=None):
        """Generalised graphical output of a Historian

        :param historian: An instance of the Historian class
        :param twindow: Amount of time to show in the plot
        :param layout: A tuple of tuples indicating how the fields should be
                plotted. For example (("T1", "T2"), ("Q1", "Q2")) indicates
                that there will be two subplots with the two indicated fields
                plotted on each.

                Note: A single plot is specified as (("T1",),) (note the commas)
        """
        import matplotlib.pyplot as plt
        from matplotlib import get_backend
        self.backend = get_backend()
        self.historian = historian
        self.twindow = twindow
        self.last_plot_update = 0
        self.last_plotted_time = 0

        if layout is None:
            layout = tuple((field,) for field in historian.columns[1:])
        self.layout = layout

        line_options = {'where': 'post', 'lw': 2, 'alpha': 0.8}
        self.lines = {}
        self.fig, self.axes = plt.subplots(len(layout), 1, figsize=(8, 1.5*len(layout)),
                                           dpi=80,
                                           sharex=True,
                                           gridspec_kw={'hspace': 0},
                                           squeeze=False)
        self.axes = self.axes[:, 0]
        values = {c: 0 for c in historian.columns}
        plt.setp([a.get_xticklabels() for a in self.axes[:-1]], visible=False)
        for axis, fields in zip(self.axes, self.layout):
            for field in fields:
                y = values[field]
                self.lines[field] = axis.step(0, y, label=field,
                                              **line_options)[0]
            axis.set_xlim(0, self.twindow)
            axis.autoscale(axis='y', tight=False)
            axis.set_ylabel(', '.join(fields))
            if len(fields) > 1:
                axis.legend()
            axis.grid()

        self.axes[-1].set_xlabel('Time / Seconds')
        plt.tight_layout()
        self.fig.canvas.draw()
        self.fig.show()
        if self.backend != 'nbAgg':
            from IPython import display
            self.display = display
            self.display.clear_output(wait=True)
            self.display.display(self.fig)

    def update(self, tnow=None):
        self.historian.update(tnow)

        minfps = 3
        maxskip = 50

        clocktime_since_refresh = time.time() - self.last_plot_update
        simtime_since_refresh = self.historian.tnow - self.last_plotted_time

        if clocktime_since_refresh <= 1/minfps and simtime_since_refresh < maxskip:
            return

        tmin = max(self.historian.tnow - self.twindow, 0)
        tmax = max(self.historian.tnow, self.twindow)
        for axis in self.axes:
            axis.set_xlim(tmin, tmax)
        data = self.historian.after(tmin)
        datadict = dict(zip(self.historian.columns, data))
        t = datadict['Time']
        for axis, fields in zip(self.axes, self.layout):
            for field in fields:
                y = datadict[field]
                self.lines[field].set_data(t, y)
            axis.relim()
            axis.autoscale_view()
        self.fig.canvas.draw()
        if self.backend != 'nbAgg':
            self.display.clear_output(wait=True)
            self.display.display(self.fig)

        self.last_plot_update = time.time()
        self.last_plotted_time = self.historian.tnow




import datetime
import tornado

from .tclab import TCLab, TCLabModel
from .historian import Historian, Plotter
from .labtime import labtime, clock

from ipywidgets import Button, Label, FloatSlider, HBox, VBox, Checkbox,\
    IntText


def actionbutton(description, action, disabled=True):
    """Return button widget with specified label and callback action."""
    button = Button(description=description, disabled=disabled)
    button.on_click(action)

    return button


def labelledvalue(label, value, units=''):
    """Return widget and HBox for label, value, units display."""
    labelwidget = Label(value=label)
    valuewidget = Label(value=str(value))
    unitwidget = Label(value=units)
    box = HBox([labelwidget, valuewidget, unitwidget])

    return valuewidget, box


def slider(label, action=None, minvalue=0, maxvalue=100, disabled=True):
    """Return slider widget for specified label and action callback."""
    sliderwidget = FloatSlider(description=label, min=minvalue, max=maxvalue)
    sliderwidget.disabled = disabled
    if action:
        sliderwidget.observe(action, names='value')

    return sliderwidget


class NotebookInteraction():
    """Base class for Notebook UI interaction controllers.

    You should inherit from this class to build new interactions.
    """
    def __init__(self):
        self.lab = None
        self.ui = None

    def update(self, t):
        """Called on a timer to update the interaction.

        t is the current simulation time. """
        raise NotImplementedError

    def connect(self, lab):
        """This is called when the interface connects to a lab

        lab is an instance of TCLab or TCLabModel
        """
        self.lab = lab
        self.lab.connected = True

    def start(self):
        """Called when the Start button is pressed"""
        raise NotImplementedError

    def stop(self):
        """Called when the Stop button is pressed"""
        raise NotImplementedError

    def disconnect(self):
        """Called when the interface disconnects from a lab"""
        self.lab.connected = False


class SimpleInteraction(NotebookInteraction):
    """Simple interaction with the TCLab

    Provides a ui with two sliders for the heaters and text boxes showing
    the temperatures.

    Notice that the class must define a "layout" property suitable to pass as
    thelayout argument to Plotter. It must also define a "sources" property
    suitable to pass to Historian. This is typically only possible after
    connecting, so in this class we define sources in .connect()

    """
    def __init__(self):
        super().__init__()

        self.layout = (('Q1', 'Q2'),
                       ('T1', 'T2'))

        # Sliders for heaters
        self.Q1widget = slider('Q1', self.action_Q1)
        self.Q2widget = slider('Q2', self.action_Q2)

        heaters = VBox([self.Q1widget, self.Q2widget])

        # Temperature display
        self.T1widget, T1box = labelledvalue('T1:', 0, '°C')
        self.T2widget, T2box = labelledvalue('T2:', 0, '°C')

        temperatures = VBox([T1box, T2box])

        self.ui = HBox([heaters, temperatures])

    def update(self, t):
        self.T1widget.value = '{:2.1f}'.format(self.lab.T1)
        self.T2widget.value = '{:2.1f}'.format(self.lab.T2)

    def connect(self, lab):
        super().connect(lab)
        self.sources = self.lab.sources

    def start(self):
        self.Q1widget.disabled = False
        self.Q2widget.disabled = False

    def stop(self):
        self.Q1widget.disabled = True
        self.Q2widget.disabled = True

    def action_Q1(self, change):
        """Change heater 1 power."""
        self.lab.Q1(change['new'])

    def action_Q2(self, change):
        """Change heater 2 power."""
        self.lab.Q2(change['new'])


class NotebookUI:
    def __init__(self, Controller=SimpleInteraction):
        self.timer = tornado.ioloop.PeriodicCallback(self.update, 1000)
        self.lab = None
        self.plotter = None
        self.historian = None
        self.seconds = 0
        self.firstsession = True

        # Model or real
        self.usemodel = Checkbox(value=False, description='Use model')
        self.usemodel.observe(self.togglemodel, names='value')
        self.speedup = slider('Speedup', minvalue=1, maxvalue=10)
        modelbox = HBox([self.usemodel, self.speedup])

        # Buttons
        self.connect = actionbutton('Connect', self.action_connect, False)
        self.start = actionbutton('Start', self.action_start)
        self.stop = actionbutton('Stop', self.action_stop)
        self.disconnect = actionbutton('Disconnect', self.action_disconnect)

        buttons = HBox([self.connect, self.start, self.stop, self.disconnect])

        # status
        self.timewidget, timebox = labelledvalue('Timestamp:', 'No data')
        self.sessionwidget, sessionbox = labelledvalue('Session:', 'No data')
        statusbox = HBox([timebox, sessionbox])

        self.controller = Controller()

        self.gui = VBox([HBox([modelbox, buttons]),
                         statusbox,
                         self.controller.ui,
                         ])

    def update(self):
        """Update GUI display."""
        self.timer.callback_time = 1000/self.speedup.value
        labtime.set_rate(self.speedup.value)

        self.timewidget.value = '{:.2f}'.format(labtime.time())
        self.controller.update(labtime.time())
        self.plotter.update(labtime.time())

    def togglemodel(self, change):
        """Speedup can only be enabled when working with the model"""
        self.speedup.disabled = not change['new']
        self.speedup.value = 1

    def action_start(self, widget):
        """Start TCLab operation."""
        if not self.firstsession:
            self.historian.new_session()
        self.firstsession = False
        self.sessionwidget.value = str(self.historian.session)

        self.start.disabled = True
        self.stop.disabled = False
        self.disconnect.disabled = True

        self.controller.start()
        self.timer.start()
        labtime.reset()
        labtime.start()

    def action_stop(self, widget):
        """Stop TCLab operation."""
        self.timer.stop()
        labtime.stop()

        self.start.disabled = False
        self.stop.disabled = True
        self.disconnect.disabled = False
        self.controller.stop()

    def action_connect(self, widget):
        """Connect to TCLab."""
        if self.usemodel.value:
            self.lab = TCLabModel()
        else:
            self.lab = TCLab()
        labtime.stop()
        labtime.reset()

        self.controller.connect(self.lab)
        self.historian = Historian(self.controller.sources)
        self.plotter = Plotter(self.historian,
                               twindow=500,
                               layout=self.controller.layout)

        self.usemodel.disabled = True
        self.connect.disabled = True
        self.start.disabled = False
        self.disconnect.disabled = False

    def action_disconnect(self, widget):
        """Disconnect TCLab."""
        self.lab.close()

        self.controller.disconnect()

        self.usemodel.disabled = False
        self.connect.disabled = False
        self.disconnect.disabled = True
        self.start.disabled = True
